/*****************************************************************
 * gavl - a general purpose audio/video processing library
 *
 * Copyright (c) 2001 - 2024 Members of the Gmerlin project
 * http://github.com/bplaum
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * *****************************************************************/



extern const uint8_t gavl_yj_8_to_y_8[256];
extern const uint8_t gavl_uvj_8_to_uv_8[256];
extern const uint16_t gavl_yj_8_to_y_16[256];
extern const uint16_t gavl_uvj_8_to_uv_16[256];
extern const uint8_t gavl_y_8_to_yj_8[256];
extern const uint8_t gavl_uv_8_to_uvj_8[256];
extern const uint16_t gavl_y_8_to_yj_16[256];
extern const float gavl_y_8_to_y_float[256];
extern const float gavl_uv_8_to_uv_float[256];
extern const float gavl_yj_8_to_y_float[256];
extern const uint8_t gavl_rgb_5_to_8[32];
extern const uint8_t gavl_rgb_6_to_8[64];
extern const uint16_t gavl_rgb_5_to_16[32];
extern const uint16_t gavl_rgb_6_to_16[64];
extern const float gavl_rgb_5_to_float[32];
extern const float gavl_rgb_6_to_float[64];
extern const int gavl_r_to_y[256];
extern const int gavl_g_to_y[256];
extern const int gavl_b_to_y[256];
extern const int gavl_r_to_u[256];
extern const int gavl_g_to_u[256];
extern const int gavl_b_to_u[256];
extern const int gavl_r_to_v[256];
extern const int gavl_g_to_v[256];
extern const int gavl_b_to_v[256];
extern const int gavl_r_to_yj[256];
extern const int gavl_g_to_yj[256];
extern const int gavl_b_to_yj[256];
extern const int gavl_r_to_uj[256];
extern const int gavl_g_to_uj[256];
extern const int gavl_b_to_uj[256];
extern const int gavl_r_to_vj[256];
extern const int gavl_g_to_vj[256];
extern const int gavl_b_to_vj[256];
extern const float gavl_r_to_y_float[256];
extern const float gavl_g_to_y_float[256];
extern const float gavl_b_to_y_float[256];
extern const float gavl_r_to_u_float[256];
extern const float gavl_g_to_u_float[256];
extern const float gavl_b_to_u_float[256];
extern const float gavl_r_to_v_float[256];
extern const float gavl_g_to_v_float[256];
extern const float gavl_b_to_v_float[256];
extern const int gavl_y_to_rgb[256];
extern const int gavl_v_to_r[256];
extern const int gavl_u_to_g[256];
extern const int gavl_v_to_g[256];
extern const int gavl_u_to_b[256];
extern const int gavl_yj_to_rgb[256];
extern const int gavl_vj_to_r[256];
extern const int gavl_uj_to_g[256];
extern const int gavl_vj_to_g[256];
extern const int gavl_uj_to_b[256];
extern const float gavl_y_to_rgb_float[256];
extern const float gavl_v_to_r_float[256];
extern const float gavl_u_to_g_float[256];
extern const float gavl_v_to_g_float[256];
extern const float gavl_u_to_b_float[256];
extern const float gavl_yj_to_rgb_float[256];
extern const float gavl_vj_to_r_float[256];
extern const float gavl_uj_to_g_float[256];
extern const float gavl_vj_to_g_float[256];
extern const float gavl_uj_to_b_float[256];
